// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test6mat.cpp
// Purpose:	Time Measurements for Rule Implementations: Materialization
// Last Change:	04.02.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Test Program 6:
//
// p3(X, Z) :- p2(X, Y), q(Y, Z).
// p2(X, Z) :- p1(X, Y), q(Y, Z).
// p1(X, Z) :- p0(X, Y), q(Y, Z).
// p0(X, Y) :- r(Y, Z).
//
// q = {(ABCDEFGHIJKLMNO-<i>, ABCDEFGHIJKLMO-<i+1>) | 1 <= i <= 5000}
// r = {(ABCDEFGHIJKLMNO-<i>, ABCDEFGHIJKLMNO-<i>)  | 1 <= i <= 5000}

#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int Q_TUPLES = 5000;

const int R_TUPLES = 5000;

const int STR_MAXSIZE = 1000;

const bool PRINT_EXTENSION = false;

// ----------------------------------------------------------------------------
// String Type (Pointer):
// ----------------------------------------------------------------------------

typedef const char *str_t;

bool str_eq(str_t s1, str_t s2)
{
	while(*s1 && *s1 == *s2) {
		s1++;
		s2++;
	}
	return (*s1=='\0' && *s2=='\0');
}

void str_cpy(char *buf, int buf_size, str_t s)
{
	while(buf_size-->0 && *s)
		*buf++ = *s++;
	if(buf_size>0)
		*buf = '\0';
	else {
		cout << "String buffer too small!\n";
		exit(1);
	}
}

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


// ----------------------------------------------------------------------------
// Predicate p0 - implemented as linked list
// ----------------------------------------------------------------------------

class p0 {
	friend class p0_cursor;
public:
	str_t col_1() const {return Col_1; }
	str_t col_2() const {return Col_2; }

	static void insert(str_t c1, str_t c2) {
		p0 *new_elem = new p0(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p0(str_t c1, str_t c2)
	{
		Col_1 = c1;
		Col_2 = c2;
		Next = 0;
	}

	str_t Col_1;
	str_t Col_2;
	p0 *Next;

	static class p0* First;
	static class p0* Last;
};

class p0* p0::First;
class p0* p0::Last;

class p0_cursor {
public:
	p0_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p0::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	str_t col_1() { return Current->col_1(); }
	str_t col_2() { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	p0*   Current;
};


// ----------------------------------------------------------------------------
// Predicate p1 - implemented as linked list
// ----------------------------------------------------------------------------

class p1 {
	friend class p1_cursor;
public:
	str_t col_1() const {return Col_1; }
	str_t col_2() const {return Col_2; }

	static void insert(str_t c1, str_t c2) {
		p1 *new_elem = new p1(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p1(str_t c1, str_t c2)
	{
		Col_1 = c1;
		Col_2 = c2;
		Next = 0;
	}

	str_t Col_1;
	str_t Col_2;
	p1 *Next;

	static class p1* First;
	static class p1* Last;
};

class p1* p1::First;
class p1* p1::Last;

class p1_cursor {
public:
	p1_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p1::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	str_t col_1() { return Current->col_1(); }
	str_t col_2() { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	p1*   Current;
};


// ----------------------------------------------------------------------------
// Predicate p2 - implemented as linked list
// ----------------------------------------------------------------------------

class p2 {
	friend class p2_cursor;
public:
	str_t col_1() const {return Col_1; }
	str_t col_2() const {return Col_2; }

	static void insert(str_t c1, str_t c2) {
		p2 *new_elem = new p2(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p2(str_t c1, str_t c2)
	{
		Col_1 = c1;
		Col_2 = c2;
		Next = 0;
	}

	str_t Col_1;
	str_t Col_2;
	p2 *Next;

	static class p2* First;
	static class p2* Last;
};

class p2* p2::First;
class p2* p2::Last;

class p2_cursor {
public:
	p2_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p2::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	str_t col_1() { return Current->col_1(); }
	str_t col_2() { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	p2*   Current;
};


// ----------------------------------------------------------------------------
// Predicate p3 - implemented as linked list
// ----------------------------------------------------------------------------

class p3 {
	friend class p3_cursor;
public:
	str_t col_1() const {return Col_1; }
	str_t col_2() const {return Col_2; }

	static void insert(str_t c1, str_t c2) {
		p3 *new_elem = new p3(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p3(str_t c1, str_t c2)
	{
		Col_1 = c1;
		Col_2 = c2;
		Next = 0;
	}

	str_t Col_1;
	str_t Col_2;
	p3 *Next;

	static class p3* First;
	static class p3* Last;
};

class p3* p3::First;
class p3* p3::Last;

class p3_cursor {
public:
	p3_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p3::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	str_t col_1() { return Current->col_1(); }
	str_t col_2() { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	p3*   Current;
};

// ----------------------------------------------------------------------------
// Predicate q - implemented as linked list
// ----------------------------------------------------------------------------

class q {
	friend class q_cursor;
public:
	str_t col_1() const {return Col_1; }
	str_t col_2() const {return Col_2; }

	static void insert(str_t c1, str_t c2) {
		q *new_elem = new q(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	q(str_t c1, str_t c2)
	{
		str_cpy(Col_1, STR_MAXSIZE, c1);
		str_cpy(Col_2, STR_MAXSIZE, c2);
		Next = 0;
	}

	char Col_1[STR_MAXSIZE];
	char Col_2[STR_MAXSIZE];
	q *Next;

	static class q* First;
	static class q* Last;
};

class q* q::First;
class q* q::Last;

class q_cursor {
public:
	q_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = q::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	str_t col_1() { return Current->col_1(); }
	str_t col_2() { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	q*   Current;
};


// ----------------------------------------------------------------------------
// Predicate r - implemented as linked list
// ----------------------------------------------------------------------------

class r {
	friend class r_cursor;
public:
	str_t col_1() const {return Col_1; }
	str_t col_2() const {return Col_2; }

	static void insert(str_t c1, str_t c2) {
		r *new_elem = new r(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	r(str_t c1, str_t c2)
	{
		str_cpy(Col_1, STR_MAXSIZE, c1);
		str_cpy(Col_2, STR_MAXSIZE, c2);
		Next = 0;
	}

	char Col_1[STR_MAXSIZE];
	char Col_2[STR_MAXSIZE];
	r *Next;

	static class r* First;
	static class r* Last;
};

class r* r::First;
class r* r::Last;

class r_cursor {
public:
	r_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = r::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	str_t col_1() { return Current->col_1(); }
	str_t col_2() { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	r*   Current;
};


//-----------------------------------------------------------------------------
// Rule Implementation: Materialization
//-----------------------------------------------------------------------------

void exec_rule4()
{
	// p0(X, Y) :- r(X, Y).
	r_cursor r_c;
	r_c.open();
	while(r_c.fetch()) {
		str_t X = r_c.col_1();
		str_t Y = r_c.col_2();
		p0::insert(X, Y);
	}
}

void exec_rule3()
{

	// p1(X, Z) :- p0(X, Y), q(Y, Z).
	p0_cursor p0_c;
	p0_c.open();
	while(p0_c.fetch()) {
		str_t X = p0_c.col_1();
		str_t Y = p0_c.col_2();
		q_cursor q_c;
		q_c.open();
		while(q_c.fetch()) {
			if(str_eq(q_c.col_1(), Y)) {
				str_t Z = q_c.col_2();
				p1::insert(X, Z);
			}
		}
	}
}

void exec_rule2()
{

	// p2(X, Z) :- p1(X, Y), q(Y, Z).
	p1_cursor p1_c;
	p1_c.open();
	while(p1_c.fetch()) {
		str_t X = p1_c.col_1();
		str_t Y = p1_c.col_2();
		q_cursor q_c;
		q_c.open();
		while(q_c.fetch()) {
			if(str_eq(q_c.col_1(), Y)) {
				str_t Z = q_c.col_2();
				p2::insert(X, Z);
			}
		}
	}
}

void exec_rule1()
{

	// p3(X, Z) :- p2(X, Y), q(Y, Z).
	p2_cursor p2_c;
	p2_c.open();
	while(p2_c.fetch()) {
		str_t X = p2_c.col_1();
		str_t Y = p2_c.col_2();
		q_cursor q_c;
		q_c.open();
		while(q_c.fetch()) {
			if(str_eq(q_c.col_1(), Y)) {
				str_t Z = q_c.col_2();
				p3::insert(X, Z);
			}
		}
	}
}

void exec_prog()
{
	// Compute p0:
	exec_rule4();

	// Compute p1:
	exec_rule3();

	// Compute p2:
	exec_rule2();

	// Compute p1:
	exec_rule1();
}

//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

// #ifdef WINDOWS
// #pragma warning(disable : 4996)
// #endif
// We use sprintf, which is safe here if MAX_STRSIZE > 21

int main()
{
	timer t;
	int num_tuples = 0;
	char buf1[STR_MAXSIZE];
	char buf2[STR_MAXSIZE];

	// Headline:
	cout << "Test of Method 1: Materialization [Program 6]\n";

	cout << "\nSize of q/r-records: " << sizeof(class q) << "\n";

	// Generation of Test Data (for EDB-predicates q and r):
	for(int i = 1; i < Q_TUPLES; i++) {
#ifdef WINDOWS
		sprintf_s(buf1, STR_MAXSIZE, "ABCDEFGHIJKLMNO-%04d", i);
		sprintf_s(buf2, STR_MAXSIZE, "ABCDEFGHIJKLMNO-%04d", i+1);
#else
		sprintf(buf1, "ABCDEFGHIJKLMNO-%04d", i);
		sprintf(buf2, "ABCDEFGHIJKLMNO-%04d", i+1);
#endif
		q::insert(buf1, buf2);
	}
	for(int i = 1; i < R_TUPLES; i++) {
#ifdef WINDOWS
		sprintf_s(buf1, STR_MAXSIZE, "ABCDEFGHIJKLMNO-%04d", i);
		sprintf_s(buf2, STR_MAXSIZE, "ABCDEFGHIJKLMNO-%04d", i);
#else
		sprintf(buf1, "ABCDEFGHIJKLMNO-%04d", i);
		sprintf(buf2, "ABCDEFGHIJKLMNO-%04d", i);
#endif
		r::insert(buf1, buf2);
	}

	// Timed Section:
	t.start();
	exec_prog();
	p3_cursor p3_c;
	p3_c.open();
	while(p3_c.fetch()) {
		num_tuples++;
		// Access both columns:
		if(str_eq(p3_c.col_1(), p3_c.col_2())) {
			cout << "Impossible!\n";
			exit(1);
		}
	}
	p3_c.close();
	t.end();

	if(PRINT_EXTENSION) {
		cout << "Extension of p3:\n\n";
		p3_c.open();
		while(p3_c.fetch()) {
			cout << "p3('" << p3_c.col_1() << "', '"
				<< p3_c.col_2() << "').\n";
		}
		p3_c.close();
	}

	cout << "Total number of tuples: " << num_tuples << "\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n\n\n";

	return 0;
}


